# -*- coding: utf-8 -*-
'''
Created on Sun Dec 20 14:49:19 2024

@author: mstep
'''

def fit_to_contents(window):
    '''Size the window to fit its contents.'''
    width = window.winfo_reqwidth()
    height = window.winfo_reqheight()
    window.geometry(f'{width}x{height}')
    window.resizable(False, False)

#%%

# tada.wav: Fanfare - Rpg by colorsCrimsonTears -- https://freesound.org/s/566203/ -- License: Creative Commons 0
# lose.wav: Sad Trombone.wav by Benboncan -- https://freesound.org/s/73581/ -- License: Attribution 4.0
# meow.wav: kitten11.wav by Department64 -- https://freesound.org/s/64006/ -- License: Creative Commons 0

import tkinter as tk
import random
import winsound

class TicTacToeApp:
    def __init__(self):
        self.game_over = False

        # Make the main interface.
        self.window = tk.Tk()
        self.window.title('tic_tac_toe')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('500x400')

        # Menus.
        self.build_menus()

        # Put the grid in a frame.
        pad = 5
        frame = tk.Frame(self.window)
        frame.pack(anchor=tk.NW, padx=pad, pady=pad)

        # Make the squares.
        self.squares = [[c for c in range(3)] for r in range(3)]
        print(self.squares)
        for r in range(3):
            for c in range(3):
                label = tk.Label(frame, text='', font=('Arial', 40, 'bold'),
                    width=4, height=2, borderwidth=3, relief=tk.SUNKEN,
                    name=f'({r}, {c})')
                label.grid(row=r, column=c, padx=pad, pady=pad)
                self.squares[r][c] = label
                label.bind('<Button>', lambda _, l=label: self.label_click(l))

        # Ensure the window is rendered.
        self.window.update_idletasks()

        # Size the window to fit the grid.
        fit_to_contents(self.window)

        # Get the default label background color.
        self.label_bg = self.squares[0][0].cget('bg')

        # Start a new game with the player being X.
        self.new_game('X')

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def build_menus(self):
        '''Build the menu bar.'''
        self.menu_bar = tk.Menu(self.window)
        self.window.config(menu=self.menu_bar)

        # Build the File menu.
        file_menu = tk.Menu(self.menu_bar, tearoff=False)
        self.menu_bar.add_cascade(label='File', menu=file_menu)

        # New Game
        file_menu.add_command(label='Play X', accelerator='Ctrl+X',
            command=lambda: self.new_game('X'))
        self.window.bind('<Control-x>', lambda _: self.new_game('X'))

        file_menu.add_command(label='Play O', accelerator='Ctrl+O',
            command=lambda: self.new_game('O'))
        self.window.bind('<Control-o>', lambda _: self.new_game('O'))

        # Exit.
        file_menu.add_separator()
        file_menu.add_command(label='Exit', command=self.kill_callback)

        # Build the Level menu.
        level_menu = tk.Menu(self.menu_bar, tearoff=False)
        self.menu_bar.add_cascade(label='Skill Level', menu=level_menu)

        self.level_var = tk.IntVar(value=1)
        for level in range(1, 4):
            level_menu.add_radiobutton(label=f'Level {level}',
                accelerator=f'Ctrl+{level}',
                variable=self.level_var, value=level)
            self.window.bind(f'<Control-Key-{level}>',
                lambda _, l=level: self.level_var.set(l))

    def kill_callback(self):
        # Stop playing any sound.
        winsound.PlaySound(None, winsound.SND_PURGE)

        # Prevent accessing widgets afert destruction.
        self.game_over = True

        # Destroy the window and stop everything.
        self.window.destroy()

    def new_game(self, player_char):
        '''Start a new game.'''
        print(f'New game: player is {player_char}')
        self.player_char = player_char
        self.computer_char = 'O' if player_char == 'X' else 'X'
        self.num_squares_taken = 0
        self.game_over = False
        self.computer_moves_next = self.computer_char == 'X'

        # Reset the board.
        self.reset_board()

        # If it's the computer's turn, let it move.
        if self.computer_moves_next:
            self.computer_move()

    def reset_board(self):
        '''Reset the board.'''
        for r in range(3):
            for c in range(3):
                self.squares[r][c]['text'] = ''
                self.squares[r][c].config(bg=self.label_bg)

    def label_click(self, label):
        # Do nothing if the square is already taken.
        if label['text'] != '': return

        # Take this square for the player.
        self.num_squares_taken += 1
        label['text'] = self.player_char

        # The computer moves next.
        # (Unless overidden by check_for_winner when we start a new game.)
        self.computer_moves_next = True

        # See if there is a winner.
        self.check_for_winner()
        
        # If the game is ending, do nothing.
        if self.game_over: return

        # If it's the computer's turn, make it move.
        if self.computer_moves_next:
            self.computer_move()

    def check_for_winner(self):
        '''See if the game is over.'''
        wins = [
            # Check columns.
            [0, 0, 1, 0],
            [0, 1, 1, 0],
            [0, 2, 1, 0],
            # Check rows.
            [0, 0, 0, 1],
            [1, 0, 0, 1],
            [2, 0, 0, 1],
            # Check diagonals.
            [0, 0,  1, 1],
            [2, 0, -1, 1],
        ]
        for win in wins:
            winner = self.row_col_winner(*win)
            if winner is not None:
                # Highlight the win and play a fanfare.
                self.highlight_win(*win)
                if winner == self.player_char:
                    winsound.PlaySound('tada.wav', winsound.SND_ASYNC)
                else:
                    winsound.PlaySound('lose.wav', winsound.SND_ASYNC)                    

                # Display a victory message and ask if we should plan again.
                if tk.messagebox.askyesno(f'{winner} wins!',
                        f'{winner} wins! Do you want to play again?'):
                    # Start a new game,
                    self.game_over = True
                    self.new_game(self.player_char)
                else:
                    # Stop.
                    self.kill_callback()
                return

        # See if it's a cat's game.
        if self.num_squares_taken == 9:
            # Cat's game.
            winsound.PlaySound('meow.wav', winsound.SND_ASYNC)
            if tk.messagebox.askyesno('Cat\'s Game',
                    'It\'s a tie! Do you want to play again?'):
                # Start a new game,
                self.game_over = True
                self.new_game(self.player_char)
            else:
                # Stop.
                self.kill_callback()
            return

    def row_col_winner(self, row, col, dr, dc):
        '''Return the row/column winner if there is one.'''
        winner = self.squares[row][col]['text']
        if winner == '':
            return None
        for i in range(1, 3):
            r = row + i * dr
            c = col + i * dc
            if self.squares[r][c]['text'] != winner:
                return None
        return winner

    def highlight_win(self, row, col, dr, dc):
        for i in range(3):
            r = row + i * dr
            c = col + i * dc
            self.squares[r][c].config(bg='yellow')

    def computer_move(self):
        # See if the board is full.
        if self.num_squares_taken == 9: return

        self.num_squares_taken += 1
        if self.level_var.get() == 1:
            self.computer_move_level_1()
        elif self.level_var.get() == 2:
            self.computer_move_level_2()
        else:
            self.computer_move_level_3()

        # The player moves next.
        # (Unless overidden by check_for_winner when we start a new game.)
        self.computer_moves_next = False

        # See if the game is over.
        self.check_for_winner()

    def computer_move_level_1(self):
        # Make a list of possible moves.
        moves = self.get_possible_moves()

        # Pick a random move.
        row, col = random.choice(moves)
        self.squares[row][col]['text'] = self.computer_char

    def get_possible_moves(self):
        '''Return a list of legal moves.'''
        moves = []
        for r in range(3):
            for c in range(3):
                if self.squares[r][c]['text'] == '':
                    moves.append([r, c])
        return moves

    def computer_move_level_2(self):
        '''
        If there is a win, take it.
        Else if there is a block, take it.
        Else move randomly.
        '''
        # Look for a win. (For more variety, you could
        # list all possible wins and pick one randomly.)
        for r in range(3):
            for c in range(3):
                if self.can_win(self.computer_char, r, c):
                    self.squares[r][c]['text'] = self.computer_char
                    return

        # Look for a block. (Again, you could randomize.)
        for r in range(3):
            for c in range(3):
                if self.can_win(self.player_char, r, c):
                    self.squares[r][c]['text'] = self.computer_char
                    return

        # Move randomly.
        self.computer_move_level_1()

    def can_win(self, char, r, c):
        '''Return True if player char can win by taking this square.'''
        # If the square ia already taken, return False.
        if self.squares[r][c]['text'] != '':
            return False

        # Temporarily take the square.
        self.squares[r][c]['text'] = char

        # See is this makes a win.
        is_win = self.row_col_winner(0, c, 1, 0)            # Column win
        is_win = is_win or self.row_col_winner(r, 0, 0, 1)  # Row win
        if r == c:                                          # Diagonal win
            is_win = is_win or self.row_col_winner(0, 0, 1, 1)
        if r + c == 2:                                      # Diagonal win
            is_win = is_win or self.row_col_winner(2, 0, -1, 1)

        # Untake the square.
        self.squares[r][c]['text'] = ''

        return is_win

    def computer_move_level_3(self):
        self.computer_move_level_1()

#%%

TicTacToeApp()
print('Done')
